# Add ability to define custom g-code macros
#
# Copyright (C) 2018  Kevin O'Connor <kevin@koconnor.net>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
import traceback, logging, math

DEFAULT_PREFIX = 'default_parameter_'


class GCodeG2G3:
    def __init__(self, config):
        self.alias = config.get_name().split()[1].upper()
        self.clockwise = True if self.alias == "G2" or self.alias == "G02" else False
        printer = config.get_printer()
        self.gcode = printer.lookup_object('gcode')
        self.gcode.register_command(self.alias, self.cmd, desc=self.cmd_desc)
        self.in_script = False
        self.kwparams = {o[len(DEFAULT_PREFIX):].upper(): config.get(o)
                         for o in config.get_prefix_options(DEFAULT_PREFIX)}

    cmd_desc = "G-Code macro"

    def cmd(self, params):
        if self.in_script:
            raise self.gcode.error(
                "Macro %s called recursively" % (self.alias,))
        self.in_script = True

        N_ARC_CORRECTION = 25
        MM_PER_ARC_SEGMENT = .4

        kwparams = dict(self.kwparams)
        kwparams.update(params)


        offset = [float(kwparams['I']), float(kwparams['J'])]

        # Radius vector from center to current location
        r_P = -offset[0]
        r_Q = -offset[1]

        # http://www.intuwiz.ru/articles/pics/g02-g03-pic4.JPG
        radius = math.hypot(r_P, r_Q)  # AC
        center_P = float(self.gcode.get_current_position()[0] - r_P)
        center_Q = float(self.gcode.get_current_position()[1] - r_Q)
        rt_X = float(float(kwparams['X']) - center_P)
        rt_Y = float(float(kwparams['Y']) - center_Q)

        logging.info(self.gcode.last_position)

        logging.info(center_P)
        logging.info(center_Q)
        logging.info(rt_X)
        logging.info(rt_Y)

        angular_travel = math.atan2(r_P * rt_Y - r_Q * rt_X, r_P * rt_X + r_Q * rt_Y)
        if angular_travel < 0:
            angular_travel += math.radians(360)
        if self.clockwise:
            angular_travel -= math.radians(360)

        if angular_travel == 0 \
                and self.gcode.base_position[0] == self.gcode.last_position[0] \
                and self.gcode.base_position[1] == self.gcode.last_position[1]:
            angular_travel = math.radians(360)

        mm_of_travel = math.fabs(radius * angular_travel)

        if mm_of_travel < 0.001:
            return

        segments = int(math.floor(mm_of_travel / MM_PER_ARC_SEGMENT))
        if segments < 1:
            segments = 1

        logging.info(str("sadsadasdsa"))
        logging.info(str(math.degrees(angular_travel)))

        raw = [0. for e in "XY"]

        theta_per_segment = angular_travel / segments
        sin_T = theta_per_segment
        cos_T = 1 - 0.5 * (theta_per_segment ** 2)

        if N_ARC_CORRECTION > 1:
            arc_recalc_count = N_ARC_CORRECTION
        logging.info(self.gcode.last_position)
        absolutecoord = self.gcode.absolutecoord
        self.gcode.absolutecoord = True
        for i in range(1, segments):
            arc_recalc_count -= 1
            if arc_recalc_count:
                r_new_Y = r_P * sin_T + r_Q * cos_T
                r_P = r_P * cos_T - r_Q * sin_T
                r_Q = r_new_Y
            else:
                if N_ARC_CORRECTION > 1:
                    arc_recalc_count = N_ARC_CORRECTION

                cos_Ti = math.cos(i * theta_per_segment)
                sin_Ti = math.sin(i * theta_per_segment)
                r_P = -offset[0] * cos_Ti + offset[1] * sin_Ti
                r_Q = -offset[0] * sin_Ti - offset[1] * cos_Ti

            raw[0] = center_P + r_P
            raw[1] = center_Q + r_Q


            self.gcode.last_position[0] = raw[0]
            self.gcode.last_position[1] = raw[1]

            self.gcode.move_with_transform(self.gcode.last_position,
                                     self.gcode.speed * self.gcode.speed_factor)
            logging.info(self.gcode.last_position)

        self.gcode.last_position[0] = float(kwparams['X'])
        self.gcode.last_position[1] = float(kwparams['Y'])
        self.gcode.move_with_transform(self.gcode.last_position,
                                       self.gcode.speed * self.gcode.speed_factor)
        logging.info(self.gcode.last_position)
        self.gcode.absolutecoord = absolutecoord
        self.in_script = False

def load_config_prefix(config):
    return GCodeG2G3(config)
